﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Algorithms.ScatterSearch {
  /// <summary>
  /// An operator that updates the solution pool using a 2-tier strategy.
  /// </summary>
  [Item("SolutionPool2TierUpdateMethod", "An operator that updates the solution pool using a 2-tier strategy.")]
  [StorableClass]
  public sealed class SolutionPool2TierUpdateMethod : SingleSuccessorOperator {
    #region Parameter properties
    public ScopeParameter CurrentScopeParameter {
      get { return (ScopeParameter)Parameters["CurrentScope"]; }
    }
    public IValueLookupParameter<SimilarityCalculator> SimilarityCalculatorParameter {
      get { return (IValueLookupParameter<SimilarityCalculator>)Parameters["SimilarityCalculator"]; }
    }
    public IValueLookupParameter<BoolValue> MaximizationParameter {
      get { return (IValueLookupParameter<BoolValue>)Parameters["Maximization"]; }
    }
    public IValueLookupParameter<BoolValue> NewSolutionsParameter {
      get { return (IValueLookupParameter<BoolValue>)Parameters["NewSolutions"]; }
    }
    public IValueLookupParameter<IntValue> NumberOfHighQualitySolutionsParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["NumberOfHighQualitySolutions"]; }
    }
    public IValueLookupParameter<IItem> QualityParameter {
      get { return (IValueLookupParameter<IItem>)Parameters["Quality"]; }
    }
    public IValueLookupParameter<IntValue> ReferenceSetSizeParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["ReferenceSetSize"]; }
    }
    public IValueLookupParameter<IItem> TargetParameter {
      get { return (IValueLookupParameter<IItem>)Parameters["Target"]; }
    }
    #endregion

    #region Properties
    private IScope CurrentScope {
      get { return CurrentScopeParameter.ActualValue; }
    }
    private SimilarityCalculator SimilarityCalculator {
      get { return SimilarityCalculatorParameter.ActualValue; }
    }
    private BoolValue Maximization {
      get { return MaximizationParameter.ActualValue; }
      set { MaximizationParameter.ActualValue = value; }
    }
    private BoolValue NewSolutions {
      get { return NewSolutionsParameter.ActualValue; }
    }
    private IntValue NumberOfHighQualitySolutions {
      get { return NumberOfHighQualitySolutionsParameter.ActualValue; }
    }
    private IItem Quality {
      get { return QualityParameter.ActualValue; }
    }
    private IntValue ReferenceSetSize {
      get { return ReferenceSetSizeParameter.ActualValue; }
      set { ReferenceSetSizeParameter.ActualValue = value; }
    }
    private IItem Target {
      get { return TargetParameter.ActualValue; }
    }
    #endregion

    [StorableConstructor]
    private SolutionPool2TierUpdateMethod(bool deserializing) : base(deserializing) { }
    private SolutionPool2TierUpdateMethod(SolutionPool2TierUpdateMethod original, Cloner cloner) : base(original, cloner) { }
    public SolutionPool2TierUpdateMethod() : base() { Initialize(); }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new SolutionPool2TierUpdateMethod(this, cloner);
    }

    private void Initialize() {
      #region Create parameters
      Parameters.Add(new ScopeParameter("CurrentScope"));
      Parameters.Add(new ValueLookupParameter<BoolValue>("Maximization"));
      Parameters.Add(new ValueLookupParameter<BoolValue>("NewSolutions"));
      Parameters.Add(new ValueLookupParameter<IntValue>("NumberOfHighQualitySolutions"));
      Parameters.Add(new ValueLookupParameter<IItem>("Quality"));
      Parameters.Add(new ValueLookupParameter<IntValue>("ReferenceSetSize"));
      Parameters.Add(new ValueLookupParameter<SimilarityCalculator>("SimilarityCalculator"));
      Parameters.Add(new ValueLookupParameter<IItem>("Target"));
      #endregion
    }

    public override IOperation Apply() {
      IScope parentsScope = new Scope("Parents");
      IScope offspringScope = new Scope("Offspring");

      // split parents and offspring
      foreach (var scope in CurrentScope.SubScopes) {
        parentsScope.SubScopes.AddRange(scope.SubScopes.Take(scope.SubScopes.Count - 1));
        offspringScope.SubScopes.AddRange(scope.SubScopes.Last().SubScopes);
      }

      var orderedParents = Maximization.Value ? parentsScope.SubScopes.OrderByDescending(x => x.Variables[QualityParameter.ActualName].Value) :
                                                parentsScope.SubScopes.OrderBy(x => x.Variables[QualityParameter.ActualName].Value);
      var orderedOffspring = Maximization.Value ? offspringScope.SubScopes.OrderByDescending(x => x.Variables[QualityParameter.ActualName].Value) :
                                                  offspringScope.SubScopes.OrderBy(x => x.Variables[QualityParameter.ActualName].Value);

      var highQualityParents = orderedParents.Take(NumberOfHighQualitySolutions.Value).ToList();
      IList<Tuple<IScope, double>> highSimilarityParents = new List<Tuple<IScope, double>>();
      foreach (var oScope in orderedParents.Skip(NumberOfHighQualitySolutions.Value)) {
        double similarity = 0.0;
        foreach (var hScope in highQualityParents) {
          similarity += SimilarityCalculator.ExecuteCalculation(oScope, hScope);
        }
        highSimilarityParents.Add(new Tuple<IScope, double>(oScope, similarity));
      }

      IList<Tuple<IScope, double>> offspring = new List<Tuple<IScope, double>>();
      foreach (var oScope in orderedOffspring) {
        double similarity = 0.0;
        foreach (var hScope in highQualityParents) {
          similarity += SimilarityCalculator.ExecuteCalculation(oScope, hScope);
        }
        offspring.Add(new Tuple<IScope, double>(oScope, similarity));
      }

      // update high quality part of the reference set
      var hasBetterQuality = Maximization.Value ? (Func<Tuple<IScope, double>, bool>)(x => { return (x.Item1.Variables[QualityParameter.ActualName].Value as DoubleValue).Value > (highQualityParents.OrderBy(y => y.Variables[QualityParameter.ActualName].Value).First().Variables[QualityParameter.ActualName].Value as DoubleValue).Value; }) :
                                                  (Func<Tuple<IScope, double>, bool>)(x => { return (x.Item1.Variables[QualityParameter.ActualName].Value as DoubleValue).Value < (highQualityParents.OrderByDescending(y => y.Variables[QualityParameter.ActualName].Value).First().Variables[QualityParameter.ActualName].Value as DoubleValue).Value; });
      if (offspring.Any(hasBetterQuality)) NewSolutions.Value = true;
      while (offspring.Any(hasBetterQuality)) { // better offspring available
        // select best offspring
        var bestChild = Maximization.Value ? offspring.OrderByDescending(x => x.Item1.Variables[QualityParameter.ActualName].Value).First() :
                                             offspring.OrderBy(x => x.Item1.Variables[QualityParameter.ActualName].Value).First();
        // select worst parent
        var worstParent = Maximization.Value ? highQualityParents.OrderByDescending(x => x.Variables[QualityParameter.ActualName].Value).Last() :
                                               highQualityParents.OrderBy(x => x.Variables[QualityParameter.ActualName].Value).Last();
        highQualityParents.Remove(worstParent);
        highQualityParents.Add(bestChild.Item1);
        offspring.Remove(bestChild);
      }

      // update diversity part of the reference set
      var hasBetterDiversity = (Func<Tuple<IScope, double>, bool>)(x => { return x.Item2 < highSimilarityParents.OrderByDescending(y => y.Item2).First().Item2; });
      if (offspring.Any(hasBetterDiversity)) NewSolutions.Value = true;
      while (offspring.Any(hasBetterDiversity)) { // better offspring available
        // select best offspring
        var bestChild = offspring.OrderBy(x => x.Item2).First();
        // select worst parent
        var worstParent = highSimilarityParents.OrderByDescending(x => x.Item2).First();
        highSimilarityParents.Remove(worstParent);
        highSimilarityParents.Add(bestChild);
        offspring.Remove(bestChild);
      }

      CurrentScope.SubScopes.Replace(highQualityParents.Concat(highSimilarityParents.Select(x => x.Item1)).ToList());

      return base.Apply();
    }
  }
}
